
#include <boost/cast.hpp>

#include "../Arch/assert.h"
#include <ctype.h>

#include "../Graphics/Graphics.h"
#include "../Graphics/GraphicsLoader.h"
#include "../Graphics/GraphicsHelpers.h"
#include "../Memory/Factory.h"
#include "../Input/InputHandler.h"
#include "../Components/Entity.h"
#include "../Components/IPosition.h"
#include "../Arch/SingletonHelpers.h"
#include "../Graphics/FontSprite.h"
#include "../Physics/Physics.h"

#include "CommandLine.h"

namespace Input_
{
	namespace { 
	const float MSG_ALPHA = 0.45f;
	const float INPUT_ALPHA = 0.85f;
	const float OUTPUT_ALPHA = 0.75f;

	const float IMG_SCALE = 0.0035f;  //0.0125

	const SDL_Color MSG_FG = {0, 0, 0, 0};
	const SDL_Color INPUT_FG = {255, 255, 255, 0};
	const SDL_Color OUTPUT_FG = {128, 128, 128, 0};

	const Math_::Vector3 MSG_POS = Math_::Vector3(-1.66f, 0.9f, -3.f);
	const Math_::Vector3 INPUT_POS = Math_::Vector3(-1.66f, 0.85f, -3.f);
	const Math_::Vector3 OUTPUT_POS = Math_::Vector3(1.25f, 0.9f, -3.f);
	}


using std::vector;
using std::string;
using Component_::Entity;
using Component_::IPosition;

CommandLine::CommandLine() 
	: inputBox("inputBox", courier, INPUT_POS, INPUT_FG, false, INPUT_ALPHA, IMG_SCALE, IMG_SCALE),
		msgBox("messageBox", arial, MSG_POS, MSG_FG, true, MSG_ALPHA, IMG_SCALE, IMG_SCALE),
		outputBox("outputBox", arial, OUTPUT_POS, OUTPUT_FG, true, OUTPUT_ALPHA, IMG_SCALE, IMG_SCALE)
{
	vector<string>& messageList = msgBox.getStrings();
	curMsg = messageList.begin();
}

void CommandLine::init()
{
	fontInit();

	inputBox.init(true);
	msgBox.init(true);
	outputBox.init(true);

	setActiveState(false);
}

CommandLine::~CommandLine()
{
	TTF_CloseFont( courier );
	TTF_CloseFont( arial );
	TTF_CloseFont( comic );
}

void CommandLine::clearInputBox(void)
{  
	inputBox.setString(" ");
}
void CommandLine::addLineToOutputBox(const std::string& str)
{
	outputBox.getStrings().push_back(str);
	outputBox.updateImage();
}
void CommandLine::addLineToMsgBox(const std::string& str)
{
	msgBox.getStrings().push_back(str);
	msgBox.updateImage();
}

void CommandLine::clearOutputBox()
{
	outputBox.getStrings().clear();
}


void CommandLine::backspaceFromInputStr(void)
{
	string& inputString = inputBox.getString();

	if(inputString.size() > 1)
	{
		inputString.erase(--inputString.end(), inputString.end());
		inputBox.updateImage();
	}
}

void CommandLine::appendToInputStr(const std::string& str)
{	
	inputBox.setString(inputBox.getString() + str);
}

void CommandLine::setInputStr(const std::string& newInputStr)
{	
	string& inputString = inputBox.getString();
	if(!newInputStr.empty())
	{
		inputString = newInputStr;
		inputBox.updateImage();
	}
}

void CommandLine::fontInit(void)
{
	assert(TTF_Init() != -1);

	courier = TTF_OpenFont( "data/fonts/cour.ttf", 12 );
	arial = TTF_OpenFont( "data/fonts/arial.ttf", 12 );
	comic = TTF_OpenFont( "data/fonts/comic.ttf", 12 );

	assert( courier && arial && comic);
}

bool CommandLine::getActiveState(void)
{  return isActive;  }

void CommandLine::setActiveState(bool as)
{  
	INPUT.setFastKeyRepeat(!as);
	isActive = as;

	inputBox.getEnt().getComp<Component_::IGraphics>()->setVisiblity(isActive);
}

void CommandLine::prevMsgToInputStr(void)
{
	vector<string>& messageList = msgBox.getStrings();
	string& inputString = inputBox.getString();

	vector<string>::iterator prevString;
	prevString = messageList.end();
	prevString--;
	int r = 0;

	if(inputString.compare(" ") && (prevString != messageList.begin()))
	{
		r = inputString.compare(*prevString);
		while((prevString != messageList.begin()) && (r!=0))
		{
			prevString--;
			r = inputString.compare(*prevString);
		}
		if(prevString != messageList.begin())
			prevString--;
	}
	
	if(prevString >= messageList.begin())
	{
		inputString = *prevString;
		inputBox.updateImage();
	}
}

void CommandLine::nextMsgToInputStr(void)
{
	vector<string>& messageList = msgBox.getStrings();
	string& inputString = inputBox.getString();

	vector<string>::iterator nextString;
	nextString = messageList.begin();
	int r = 0;

	if(inputString.compare(" "))
	{
		r = inputString.compare(*nextString);
		while( r!=0 && (nextString != messageList.end()) )
		{
			nextString++;
			if(nextString != messageList.end())
				r = inputString.compare(*nextString);
		}
		if(nextString != messageList.end())
			nextString++;
	}
	
	if(nextString != messageList.end())
	{
		inputString = *nextString;
		inputBox.updateImage();
	}
}

void CommandLine::addToMsgBox(void)
{
	vector<string>& messageList = msgBox.getStrings();
	string& inputString = inputBox.getString();

	if( (messageList.size() < 7) && inputString.compare(" ") && inputString.compare(*curMsg) )
		messageList.push_back(inputString);

	else if( (messageList.size() == 7) && inputString.compare(" ") && inputString.compare(*curMsg) )
	{
		for(int i = 0; i < 6; i++)
			messageList[i] = messageList[i+1];

		messageList[6] = inputString;
	}

	curMsg = messageList.end();
	--curMsg;
	clearInputBox();
	msgBox.updateImage();
	parseLine();
}

string subString(std::string::iterator& begin, std::string::iterator& end)
{
	string ret;
	string::const_iterator b = begin, e = end;

	while(b != e)
	{
		ret.push_back(*b);
		b++;
	}

	return ret;
}

void CommandLine::parseLine(void)
{
	string line, command, args;
	string::iterator marker, deleter, editor;
	vector<string> argList;

	//get the current command
	line = *curMsg;

	//erase the space at the beginning of the line
	deleter = line.begin();
	line.erase(deleter);

	//get the primary command substring
	marker = line.begin();
	editor = marker;
	while( (editor != line.end()) && (*editor != '-') )
		editor++;
	command = line.substr(0, static_cast<int>(editor-marker));
	if(command.empty())
		command = "error";
	marker = editor;
	editor = line.end();

	//check for and delete end space on command
	deleter = command.end();
	deleter--;
	if(*deleter == ' ')
		command.erase(deleter);

	//get argument string
	if(editor != marker)
		args = subString(marker, editor);

	//peel arguments off string into list;
	while(!args.empty())
	{
		//delete the first tick
		deleter = args.begin();
		args.erase(deleter);

		marker = args.begin();  //mark the begining of string
		editor = marker;

		//find the beginning of the next argument or end of string
		while( (editor != args.end()) && (*editor != '-') )
			editor++;

		if(editor != args.end())
		{
			//erase the leading space
			editor--;
			args.erase(editor);
		}

		//push back the substring into the list
		argList.push_back(subString(marker, editor));

		//remove the argument from the string
		args.erase(marker, editor);
	}

	//Act on the command:
	if(!command.compare("spawn") && !argList.empty())
	{
		//spawn the thing given by the first argument
		Id_::TypeId_t objectType = FACTORY.getTypeIdByName(argList[0]);

		if(objectType != Id_::INVALID_TYPE)
		{
			Component_::IPosition *pos = boost::polymorphic_downcast<Component_::Entity*>
				(FACTORY.createDynObj(objectType))->getComp<Component_::IPosition>();

			pos->setPosition(GRAPHICS.getCamera().getPosition());
		}
		else
		{
			curMsg->append(" <Invalid Arguments>");
			msgBox.updateImage();
		}
	}
	else if(!command.compare("debugControl") && !argList.empty())
	{
		if(!argList[0].compare("on"))
			INPUT.setState(DEBUG_CONTROL, true);
		else if(!argList[0].compare("off"))
			INPUT.setState(DEBUG_CONTROL, false);
		else
		{
			curMsg->append(" <Invalid Arguments>");
			msgBox.updateImage();
		}

	}
	else if(!command.compare("setGrav") && !argList.empty())
	{
		float grav = static_cast<float>(strtod(argList[0].c_str(), NULL));
		PHYSICS.getUpdater().setGravityMagnitude(grav);
	}
	else if(!command.compare("angleSnap") && !argList.empty())
	{
		if(!argList[0].compare("on"))
		{
			INPUT.setState(ANGLE_SNAP, true);
			INPUT.setRotationAngles(Math_::Vector2(0.0, 0.0));
		}
		else if(!argList[0].compare("off"))
			INPUT.setState(ANGLE_SNAP, false);
		else
		{
			curMsg->append(" <Invalid Arguments>");
			msgBox.updateImage();
		}
	}
	else  //append incalid command marker
	{
		if( curMsg->length() > 17 )  //we need to check if this is a degenerate case already marked
		{
			editor = curMsg->end();  //so get the substring defined by the last 17 chars
			marker = editor - 17;
			if(subString(marker, editor).compare("<Invalid Command>"))  //and if it desn't have the IC marker, then add it
			{
				curMsg->append(" <Invalid Command>");
				msgBox.updateImage();
			}
		}
		else  //it is too short to already have the IC marker, so append it now
		{
			curMsg->append(" <Invalid Command>");
			msgBox.updateImage();
		}
	}
}

void CommandLine::readToInputStr(unsigned int c, bool state)
{
	std::string is;
	switch(c)
	{
		case SDLK_a:
		case SDLK_b:
		case SDLK_c:
		case SDLK_d:
		case SDLK_e:
		case SDLK_f:
		case SDLK_g:
		case SDLK_h:
		case SDLK_i:
		case SDLK_j:
		case SDLK_k:
		case SDLK_l:
		case SDLK_m:
		case SDLK_n:
		case SDLK_o:
		case SDLK_p:
		case SDLK_q:
		case SDLK_r:
		case SDLK_s:
		case SDLK_t:
		case SDLK_u:
		case SDLK_v:
		case SDLK_w:
		case SDLK_x:
		case SDLK_y:
		case SDLK_z:
		case SDLK_0:
		case SDLK_1:
		case SDLK_2:
		case SDLK_3:
		case SDLK_4:
		case SDLK_5:
		case SDLK_6:
		case SDLK_7:
		case SDLK_8:
		case SDLK_9:
		case SDLK_MINUS:
		case SDLK_PERIOD:
		case SDLK_COMMA:
		case SDLK_SPACE:
			if(state)
			{
			  if(INPUT.shiftState())
				is = static_cast<char>(toupper(c));
			  else
				is = static_cast<char>(c);
			  appendToInputStr(is);
			}
		  break;

		case SDLK_LSHIFT:
		case SDLK_RSHIFT:
			INPUT.setState(SHIFT_HOLD, state);
			break;

		case SDLK_RETURN:
			if(state)
			  addToMsgBox();
			break;

		case SDLK_BACKSPACE:
		  if(state)
			backspaceFromInputStr();
		  break;

		case SDLK_UP:
		  if(state)
		    prevMsgToInputStr();
		  break;

		case SDLK_DOWN:
		  if(state)
		    nextMsgToInputStr();
		  break;

		case SDLK_ESCAPE:
			setActiveState(false);
		  break;

		default:
		  break;
	}
}

}
//END OF Input_
