#include "InputHandler.h"
#include "../Math/Math.h"
#include "../Graphics/Graphics.h"
#include "../Arch/Id.h"
#include "../Memory/Factory.h"
#include "../Arch/SingletonHelpers.h"
#include "../Components/Entity.h"
#include "../Components/IPosition.h"
#include "../Components/IPositionHelpers.h"
#include "../Components/IVelocity.h"
#include "../Components/ComponentTypes.h"
#include "../Time/Timer.h"
#include "../Physics/Physics.h"
#include "../Components/ISoundComp.h"
#include "../Sound/Sound.h"
#include "../GameLogic/JohnnyGameLogic.h"

namespace Input_
{
using std::abs;
using Math_::Vector3;
namespace {
	const InputBits TURNING = InputBits(TURN_LEFT) |= InputBits(TURN_RIGHT) |= 
										InputBits(TURN_UP) |= InputBits(TURN_DOWN);

	std::map<const SDLKey, InputBit> keyToInput;
	std::map<const SDLKey, WRInputBit> WRkeyToInput;

	void initializeInputMap()
	{
		keyToInput[SDLK_w] = RUN;
		keyToInput[SDLK_s] = BACKPEDDLE;

		// TIM CODE TEST
		keyToInput[SDLK_SPACE] = JUMP;

		keyToInput[SDLK_a] = TURN_LEFT;
		keyToInput[SDLK_d] = TURN_RIGHT;
		keyToInput[SDLK_t] = TURN_UP;
		keyToInput[SDLK_g] = TURN_DOWN;

		keyToInput[SDLK_q] = STRAFE_LEFT;
		keyToInput[SDLK_e] = STRAFE_RIGHT;
		keyToInput[SDLK_x] = STRAFE_UP;
		keyToInput[SDLK_c] = STRAFE_DOWN;

		keyToInput[SDLK_LSHIFT] = STRAFE_DOWN;
		keyToInput[SDLK_RSHIFT] = STRAFE_DOWN;
		keyToInput[SDLK_LCTRL] = CONTROL_HOLD;
		keyToInput[SDLK_RCTRL] = CONTROL_HOLD;

		keyToInput[SDLK_BACKQUOTE] = CONSOLE_TOGGLE;
		keyToInput[SDLK_ESCAPE] = QUIT;

		keyToInput[SDLK_i] = TRANS_Z_F;
		keyToInput[SDLK_k] = TRANS_Z_B;
		keyToInput[SDLK_j] = TRANS_X_L;
		keyToInput[SDLK_l] = TRANS_X_R;
		keyToInput[SDLK_u] = TRANS_Y_U;
		keyToInput[SDLK_o] = TRANS_Y_D;

		keyToInput[SDLK_KP4] = ROT_Y_L;
		keyToInput[SDLK_KP6] = ROT_Y_R;
		keyToInput[SDLK_KP8] = ROT_X_F;
		keyToInput[SDLK_KP6] = ROT_Y_R;
		keyToInput[SDLK_KP5] = ROT_X_B;
		keyToInput[SDLK_KP7] = ROT_Z_L;
		keyToInput[SDLK_KP9] = ROT_Z_R;	
	}

	void initializeWRInputMap()
	{
		WRkeyToInput[SDLK_a] = WORLD_TURN_LEFT;
		WRkeyToInput[SDLK_d] = WORLD_TURN_RIGHT;
		WRkeyToInput[SDLK_t] = WORLD_TURN_UP;
		WRkeyToInput[SDLK_g] = WORLD_TURN_DOWN;
	}

	InputBit getState(const SDLKey key)
	{
		if(keyToInput.find(key) != keyToInput.end())
			return keyToInput.find(key)->second;
		else
			return INVALID_STATE;
	}

	WRInputBit getWRState(const SDLKey key)
	{
		if(WRkeyToInput.find(key) != WRkeyToInput.end())
			return WRkeyToInput.find(key)->second;
		else
			return WORLD_INVALID_STATE;
	}
}

InputHandler::InputHandler() : selectedObject(Id_::INVALID_OBJ)
{
	setFastKeyRepeat(true); 
	clearStates();
	initializeInputMap();
	initializeWRInputMap();

	lockMouse = false;
	setState(ANGLE_SNAP, true);
}

InputHandler::~InputHandler()
{  }

void InputHandler::setJohnny(Component_::Entity* newJohn)
{
	johnny = newJohn;
}

const Component_::Entity* InputHandler::getjohnny(void)
{
	return johnny;
}

bool InputHandler::clearStates(void)
{  
	inputState.reset();
	return true;
}

void InputHandler::setState(InputBit input, bool state)
{
	inputState[input] = state;
}

void InputHandler::setWRState(WRInputBit WRinput, bool state)
{
	WRinputState[WRinput] = state;
}

void InputHandler::setFastKeyRepeat(bool fast)
{
	if(fast)
		SDL_EnableKeyRepeat(20, 15);
	else
		SDL_EnableKeyRepeat(1000, 500);
}

void InputHandler::setKeyState(SDLKey key, bool state)
 {  
	Component_::IPosition *pos = johnny->getComp<Component_::IPosition>();
	Math_::Vector4 Y = -pos->getOrientation().getColumn(1);

	switch(key)
	{
		case SDLK_r:
			GRAPHICS.getCamera().camReset();
			break;

		case SDLK_p:
			pos->setOrientation(Math_::Matrix4::IDENTITY);
			pos->setPosition(GAME_LOGIC.getPlayerStartPos());
			pos->setPreviousPosition(GAME_LOGIC.getPlayerStartPos());
			PHYSICS.getUpdater().setGravityDirection(Math_::Vector3(Y.getX(), Y.getY(), Y.getZ()));
			GRAPHICS.getCamera().camReset();
			break;

		case SDLK_F9:
			setMouseLock( true );
			break;

		case SDLK_F10:
			setMouseLock( false );
			break;

		default:
			if(worldRotateState() && debugControlState())
				WRinputState[getWRState(key)] = state;
			else
				inputState[getState(key)] = state;			
			break;
	}
}

void InputHandler::translateObject(const Math_::Vector3& disp) const
{
	Component_::IPosition *pos = FACTORY.getObj<Component_::Entity>(selectedObject)->getComp<Component_::IPosition>();

	// view->world rotation transform
	Math_::Vector3 v3 = getInnerMatrix(GRAPHICS.getCamera().getViewFrame()).getTranspose() * disp;  // CHECK

	pos->setPosition( pos->getPosition() + v3);
}

void InputHandler::setSelectionFreeze(bool freeze)
{
	if(FACTORY.getObj<Component_::Entity>(selectedObject)->hasComp(Component_::VELOCITY_COMPONENT))
	{
		Component_::IVelocity *vel = FACTORY.getObj<Component_::Entity>(selectedObject)->getComp<Component_::IVelocity>();
		
		vel->setStatic( freeze );
		if( !vel->isStatic() )
		{
			vel->setVelocity( Vector3::ZERO );
		}
	}
}

void InputHandler::setSelectionVelocity( const Math_::Vector2 &vel )
{
	if(!FACTORY.getObj<Component_::Entity>(selectedObject)->hasComp(Component_::VELOCITY_COMPONENT))
		return;

	const float VELOCITY_SCALE = 20.f;
	Component_::IVelocity *velComp = FACTORY.getObj<Component_::Entity>(selectedObject)->getComp<Component_::IVelocity>();
	
	if( velComp->isStatic() )
		return;

	Vector3 camVecX = CAMERA.getCameraVec( Vector3(1.f,0.f,0.f) );
	Vector3 camVecY = CAMERA.getCameraVec( Vector3(0.f,-1.f,0.f) );

	velComp->setVelocity( VELOCITY_SCALE*(camVecX*vel.getX() + camVecY*vel.getY()) );
}

bool InputHandler::isMouseLocked() const
{
	return lockMouse;
}

void InputHandler::setMouseLock( bool lock )
{
	lockMouse = lock;
}

const Math_::Vector2 &InputHandler::getMotionVelocity() const
{
	return motionVelocity;
}

void InputHandler::setMotionVelocity( const Math_::Vector2 &vel )
{
	motionVelocity = vel;
}

void InputHandler::rotateObject(float phi, float theta, float rho) const
{
	Component_::Entity& object = *FACTORY.getObj<Component_::Entity>(selectedObject);
	Component_::IPosition *pos = object.getComp<Component_::IPosition>();
	Math_::Matrix4 mx, my, mz, mf;

	if(abs(phi) > 0.01)
	{
		mx.set(1,1, cos(phi));
		mx.set(1,2, -sin(phi));
		mx.set(2,1, sin(phi));
		mx.set(2,2, cos(phi));
	}

	if(abs(theta) > 0.01)
	{
		my.set(0,0, cos(theta));
		my.set(0,2, sin(theta));
		my.set(2,0, -sin(theta));
		my.set(2,2, cos(theta));
	}

	if(abs(rho) > 0.01)
	{
		mz.set(0,0, cos(rho));
		mz.set(0,1, -sin(rho));
		mz.set(1,0, sin(rho));
		mz.set(1,1, cos(rho));
	}

	mf = mx*my*mz;

	pos->setOrientation(pos->getOrientation() * mf);
}
	
//non-const accessor for the selected object ID
const Id_::ObjId_t InputHandler::getSelectedObject(void) const
{  return selectedObject;  }

//setter for the selected object
void InputHandler::setSelectedObject(Id_::ObjId_t objectID)
{  selectedObject = objectID;  }

void InputHandler::rectifyRotationAngle(void)
{
	float increment = (2.f*Math_::pi)/24.f;

	float spinInc = rotationAngles.getX()/increment;
	float spinDif = spinInc - static_cast<int>(spinInc);
	float pitchInc = rotationAngles.getY()/increment;
	float pitchDif = pitchInc - static_cast<int>(pitchInc);

	float spinOffset = -spinDif * increment;
	float pitchOffset = -pitchDif * increment;

	if( abs(spinDif) > 0.5f )
		spinOffset = (1-spinDif) * increment;

	if( abs(pitchDif) > 0.5f )
		pitchOffset = (1-pitchDif) * increment;

	Position_::rotateBarrel(*johnny->getComp<Component_::IPosition>(), spinOffset);
	rotationAngles.setX(rotationAngles.getX() + spinOffset);

	Position_::rotateUp(*johnny->getComp<Component_::IPosition>(), pitchOffset);
	rotationAngles.setY(rotationAngles.getY() + pitchOffset);
}

//returns the current state of the shift key
bool InputHandler::shiftState(void)
{  return inputState.test(SHIFT_HOLD);  }

//returns the current state of the control key
bool InputHandler::controlState(void)
{  return inputState.test(CONTROL_HOLD);  }

//returns whether the game is in world rotation mode
bool InputHandler::worldRotateState(void)
{  return inputState.test(WORLD_ROTATE);  }

bool InputHandler::worldPreviewRotateState(void)
{  return inputState.test(WORLD_ROTATE_PREVIEW);  }

bool InputHandler::debugControlState(void)
{  return inputState.test(DEBUG_CONTROL);  }

bool InputHandler::angleSnapState(void)
{  return inputState.test(ANGLE_SNAP);  }

void InputHandler::setRotationAngles(const Math_::Vector2& rvec)
{  rotationAngles = rvec;  }

//returns whether the game is in object translation mode
bool InputHandler::objectTranslatieState(void)
{  return inputState.test(OBJECT_TRANSLATE);  }

//returns whether the game is in object rotation mode
bool InputHandler::objectRotateState(void)
{  return inputState.test(OBJECT_ROTATE);  }

//update the player based on the current input states
void InputHandler::update(Camera_::Camera& cam) const
{
	Math_::Vector3 v;
	float x = 0.f, y = 0.f, z = 0.f;
	const float OBJECTSPEED = .25f;
	//const float PLAYERSPEED = 350.f;

	johnnyMove = Math_::Vector3::ZERO;

	if(inputState[QUIT])
		if(COMLINE.getActiveState())
			COMLINE.setActiveState(false);
		else
			GAME_LOGIC.setGameOver();

	if(inputState[CONSOLE_TOGGLE])
	{
		COMLINE.setActiveState(!COMLINE.getActiveState());
		const_cast<InputBits&>(inputState)[CONSOLE_TOGGLE] = false;
	}

	if(inputState[RUN] && !inputState[WORLD_ROTATE_PREVIEW])
	{
		v += Math_::Vector3(0.f,0.f,1.f);
	}	
	else if(inputState[BACKPEDDLE] && !inputState[WORLD_ROTATE_PREVIEW])
	{
		v += Math_::Vector3(0.f,0.f,-.5f);
	}

	// THIS IS TIM CODE, HE'S JUST TRYING TO HELP SO DON'T KILL HIM!!! Oh, and by the way, this is jumping
	if(inputState[JUMP] && !inputState[WORLD_ROTATE_PREVIEW])
	{
		if(PHYSICS.getUpdater().isPlayerOnGround())
		{
			(*johnny->getComp<Component_::IVelocity>()).setVelocity( 
				(*johnny->getComp<Component_::IVelocity>()).getVelocity() - 300.f*PHYSICS.getUpdater().getGravityDirection() );

			(*johnny->getComp<Component_::ISoundComp>()).playSound(Sound_::EVENTID_JOHNNY_SFX_JUMP);
		}
	}

	if(inputState[WORLD_ROTATE] )
	{
		Math_::Vector2 movement = motionVelocity / 80.f;

		rotationAngles += movement;

		movement.setX( abs(movement.getX()) );
		movement.setY( abs(movement.getY()) );
		if(movement == Math_::Vector2::ZERO)
		{
			movement.setX(TIMER.getTimeDelta() * 2.f);
			movement.setY(TIMER.getTimeDelta() * 2.f);
		}

		if(WRinputState[WORLD_TURN_LEFT])
		{
			Position_::rotateBarrel(*johnny->getComp<Component_::IPosition>(), movement.getX());
			cam.updateWorld();
		}

		if(WRinputState[WORLD_TURN_RIGHT])
		{
			Position_::rotateBarrel(*johnny->getComp<Component_::IPosition>(), -movement.getX());
			cam.updateWorld();
		}

		if(WRinputState[WORLD_TURN_UP])
		{
			Position_::rotateUp(*johnny->getComp<Component_::IPosition>(), -movement.getY());
			cam.updateWorld();
		}

		if(WRinputState[WORLD_TURN_DOWN])
		{
			Position_::rotateUp(*johnny->getComp<Component_::IPosition>(), movement.getY());
			cam.updateWorld();
		}
	}

	if(inputState[TURN_LEFT])
	{
		Position_::rotateRight(*johnny->getComp<Component_::IPosition>(), -TIMER.getTimeDelta() * 2.f);
		cam.updateWorld();
	}

	if(inputState[TURN_RIGHT])
	{
		Position_::rotateRight(*johnny->getComp<Component_::IPosition>(), TIMER.getTimeDelta() * 2.f);
		cam.updateWorld();
	}

	if(inputState[TURN_UP])
		cam.lookUp(-TIMER.getTimeDelta() * 2.f);

	if(inputState[TURN_DOWN])
		cam.lookUp(TIMER.getTimeDelta() * 2.f);

	if(inputState[STRAFE_LEFT])
	{
		v += Math_::Vector3(1.f,0.f,0.f);
		//v.setX(PLAYERSPEED);
	}

	if(inputState[STRAFE_RIGHT])
	{
		v += Math_::Vector3(-1.f,0.f,0.f);
		//v.setX(-PLAYERSPEED);
	}

	if(inputState[STRAFE_UP])
		v += Math_::Vector3(0.f,1.f,0.f);
		//v.setY(PLAYERSPEED);

	if(inputState[STRAFE_DOWN])
		v += Math_::Vector3(0.f,-1.f,0.f);
		//v.setY(-PLAYERSPEED);

	if(johnny && v != v.ZERO)
	{
		v.normalize();
		//v *= PLAYERSPEED;

		//Component_::IVelocity *vel = johnny->getComp<Component_::IVelocity>();
		Component_::IPosition *pos = johnny->getComp<Component_::IPosition>();

		johnnyMove = Math_::getInnerMatrix(pos->getOrientation())*v;
		//vel->addVelocity( Math_::getInnerMatrix(pos->getOrientation())*v );

		//Position_::move(*pos, v);
		cam.updateWorld();
	}
	else
		cam.move(v);
	
	
	v.setToZero();

	if(inputState[TRANS_Z_F])
		v.setZ(-OBJECTSPEED);

	if(inputState[TRANS_Z_B])
		v.setZ(OBJECTSPEED);

	if(inputState[TRANS_X_L])
		v.setX(-OBJECTSPEED);

	if(inputState[TRANS_X_R])
		v.setX(OBJECTSPEED);

	if(inputState[TRANS_Y_U])
		v.setY(OBJECTSPEED);

	if(inputState[TRANS_Y_D])
		v.setY(-OBJECTSPEED);

	if(inputState[ROT_X_F])
		x = -1.f/40.f;

	if(inputState[ROT_X_B])
		x = 1.f/40.f;

	if(inputState[ROT_Y_L])
		y = -1.f/40.f;

	if(inputState[ROT_Y_R])
		y = 1.f/40.f;

	if(inputState[ROT_Z_L])
		z = 1.f/40.f;

	if(inputState[ROT_Z_R])
		z = -1.f/40.f;

	if( selectedObject != Id_::INVALID_OBJ )
	{
		translateObject(v);
		rotateObject(x, y, z);

		/*
		// DELETEME - SOUND HAX
		if( v.length() != 0.f )
		{
			Component_::ISoundComp *sound = FACTORY.getObj<Component_::Entity>(selectedObject)->getComp<Component_::ISoundComp>();
			sound->playSound(Sound_::EVENT_JOHNNY_SFX_IMPACT).setParameter(Sound_::EVENTPARAMETER_JOHNNY_SFX_IMPACT_IMPULSE, 500.f);
		}
		*/
	}
}


//this is the main input fetcher and updater
void inputUpdate(SDL_Event* input, InputHandler& handler)
{
	Camera_::Camera& cam = Graphics_::Graphics::getInstance().getCamera();

	handler.setMotionVelocity( Math_::Vector2(0.f, 0.f) );
	handler.setWRState(WORLD_TURN_RIGHT, false);
	handler.setWRState(WORLD_TURN_LEFT, false);
	handler.setWRState(WORLD_TURN_UP, false);
	handler.setWRState(WORLD_TURN_DOWN, false);

	while(SDL_PollEvent(input))
	{
		if(GRAPHICS.splashState())  //SPLASH-HACK
		{
			Component_::ISoundComp &sound = *(INPUT.getjohnny())->getComp<Component_::ISoundComp>();
			sound.playSound( Sound_::EVENTID_JOHNNY_MUSIC_MENUMUSIC1 );
			if( (input->type == SDL_KEYDOWN || input->type == SDL_KEYUP) &&
			( getState(input->key.keysym.sym) == CONSOLE_TOGGLE || !COMLINE.getActiveState()) )
			{
				handler.setKeyState( input->key.keysym.sym, input->type == SDL_KEYDOWN );
			}
			else if(input->type == SDL_MOUSEBUTTONDOWN)
			{
				switch(input->button.button)
				{
					case 1:
						GRAPHICS.setSplash(false);
						GRAPHICS.setMenu(true);
				}
			}
		}
		else if(GRAPHICS.menuState())
		{
			if( (input->type == SDL_KEYDOWN || input->type == SDL_KEYUP) &&
			( getState(input->key.keysym.sym) == CONSOLE_TOGGLE || !COMLINE.getActiveState()) )
			{
				handler.setKeyState( input->key.keysym.sym, input->type == SDL_KEYDOWN );
			}
			else if(input->type == SDL_MOUSEBUTTONDOWN)
			{
				switch(input->button.button)
				{
					case 1:
						if(input->motion.x <= 408 && input->motion.x >= 165 &&
						   input->motion.y >= 430 && input->motion.y <= 640)
						{
							Component_::ISoundComp &sound = *(INPUT.getjohnny())->getComp<Component_::ISoundComp>();
							sound.playSound( Sound_::EVENTID_JOHNNY_MUSIC_AMBIENT );
							GRAPHICS.setMenu(false);
							GRAPHICS.setGame(true);
							SDL_WM_GrabInput(SDL_GRAB_ON); // this STEALS ALL INPUT on linux (read "tyler can't debug")
							SDL_ShowCursor(SDL_DISABLE);
						}
						else if(input->motion.x <= 790 && input->motion.x >= 540 &&
								input->motion.y >= 430 && input->motion.y <= 640)
							exit(0);
				}
			}
		}
		else if( (input->type == SDL_KEYDOWN || input->type == SDL_KEYUP) &&
			( getState(input->key.keysym.sym) == CONSOLE_TOGGLE || !COMLINE.getActiveState()) )
		{
			handler.setKeyState( input->key.keysym.sym, input->type == SDL_KEYDOWN );
		}
		else if((input->type == SDL_KEYDOWN || input->type == SDL_KEYUP) && COMLINE.getActiveState())
		{
			COMLINE.readToInputStr( input->key.keysym.sym, input->type == SDL_KEYDOWN );
		}
		else if (input->type == SDL_MOUSEMOTION && 
			(input->motion.xrel != input->motion.x) &&
			(input->motion.yrel != input->motion.y))
		{
			handler.setMotionVelocity( Math_::Vector2( static_cast<float>(input->motion.xrel),
																	static_cast<float>(input->motion.yrel)) );

			if(handler.objectTranslatieState())  //when translating an object:
				handler.translateObject(Math_::Vector3(
													static_cast<float>(input->motion.xrel),
													static_cast<float>(-1.f*input->motion.yrel), 0.f));

			else if(handler.objectRotateState())  //when rotating objects
				handler.rotateObject(input->motion.yrel / 80.f, input->motion.xrel / 80.f, 0.0f);

			else
			{
				if(!handler.worldRotateState())  //rotating the camera:
				{
					cam.lookRight( -input->motion.xrel / 80.f );
					cam.lookUp( -input->motion.yrel / 80.f);
				}
				else // world rotation (mouse)
				{
					if(input->motion.xrel > 0)
					{
						handler.setWRState(WORLD_TURN_LEFT, false);
						handler.setWRState(WORLD_TURN_RIGHT, true);
					}
					else if(input->motion.xrel < 0)
					{
						handler.setWRState(WORLD_TURN_LEFT, true);
						handler.setWRState(WORLD_TURN_RIGHT, false);
					}

					
					if(input->motion.yrel > 0)
					{
						handler.setWRState(WORLD_TURN_DOWN, true);
						handler.setWRState(WORLD_TURN_UP, false);
					}
					else if(input->motion.yrel < 0)
					{
						handler.setWRState(WORLD_TURN_DOWN, false);
						handler.setWRState(WORLD_TURN_UP, true);
					}
				}
			}
		}
		else if(input->type == SDL_MOUSEBUTTONDOWN)
		{
			switch(input->button.button)
			{
				case 1:
				  if(INPUT.worldRotateState())
					  INPUT.setState(WORLD_ROTATE_PREVIEW, true);
				  else
				  {
					  handler.setSelectedObject( GRAPHICS.pick(Math_::Vector2(input->motion.x, input->motion.y)) );

					  if(handler.getSelectedObject() == Id_::INVALID_OBJ)
					  {
						  handler.setSelectedObject(Id_::INVALID_OBJ);
						  handler.setState(OBJECT_TRANSLATE, false);
						  handler.setState(OBJECT_ROTATE, false);
						  break;
					  }

					  handler.setSelectionFreeze(true);

					  if(handler.controlState())
						  handler.setState(OBJECT_ROTATE, true);
					  else
						  handler.setState(OBJECT_TRANSLATE, true);
				  }

				  break;

				case 2:
					Position_::resetToWorld(*handler.getjohnny()->getComp<Component_::IPosition>());
				  break;

				case 3:
					handler.setState(WORLD_ROTATE, true);
				    break;

				case 4:
					if(handler.objectTranslatieState())
						handler.translateObject(Math_::Vector3(0, 0, -10));
					else if(handler.objectRotateState())  //when rotating objects
						handler.rotateObject(0.f, 0.f, 1.f/20.f);
					else if(handler.controlState())
						GRAPHICS.getCamera().zoom(.1f);
					break;

				case 5:
					if(handler.objectTranslatieState())
					{
						/*Math_::Vector3 toEye = GRAPHICS.getCamera().getPosition() 
							- (FACTORY.getObj<Component_::Entity>(handler.getSelectedObject())->getComp<Component_::IPosition>())->getPosition();
						
						handler.translateObject(Math_::Vector3(toEye.getX()/20.f, toEye.getY()/20.f, toEye.getZ()/20.f));*/
						handler.translateObject(Math_::Vector3(0, 0, 10));
					}
					else if(handler.objectRotateState())  //when rotating objects
						handler.rotateObject(0.f, 0.f, -1.f/20.f);
					else if(handler.controlState())
						GRAPHICS.getCamera().zoom(-.1f);
					break;

				default:
                  break;
			}
		}
		else if(input->type == SDL_MOUSEBUTTONUP)
		{
			switch(input->button.button)
			{
				case 1:
					if(INPUT.worldRotateState())
					{
						INPUT.setState(WORLD_ROTATE_PREVIEW, false);

						Component_::IPosition *pos = INPUT.getjohnny()->getComp<Component_::IPosition>();
						Math_::Vector4 Y = -pos->getOrientation().getColumn(1);
						PHYSICS.getUpdater().setGravityDirection(Math_::Vector3(Y.getX(), Y.getY(), Y.getZ()));

						if(INPUT.angleSnapState())
							INPUT.rectifyRotationAngle();
					}
					else
					{
						handler.setState(OBJECT_TRANSLATE, false);
						handler.setState(OBJECT_ROTATE, false);

						if(handler.getSelectedObject() != Id_::INVALID_OBJ)
						{
							handler.setSelectionFreeze(false);
							handler.setSelectionVelocity( handler.getMotionVelocity() );
						}
					}
					break;

				case 3:
					{
						handler.setState(WORLD_ROTATE, false);
						handler.setState(WORLD_ROTATE_PREVIEW, false);
						handler.setWRState(WORLD_TURN_RIGHT, false);
						handler.setWRState(WORLD_TURN_LEFT, false);
						handler.setWRState(WORLD_TURN_UP, false);
						handler.setWRState(WORLD_TURN_DOWN, false);

						Component_::IPosition *pos = INPUT.getjohnny()->getComp<Component_::IPosition>();
						Math_::Vector4 Y = -pos->getOrientation().getColumn(1);
						PHYSICS.getUpdater().setGravityDirection(Math_::Vector3(Y.getX(), Y.getY(), Y.getZ()));

						if(INPUT.angleSnapState())
							INPUT.rectifyRotationAngle();
					}

					break;

				default:
                  break;
			}
		}
		if( handler.isMouseLocked() )
		{
			// FIXME - DIRTY HAX YUCKY POO POO
			// This is so that I don't go insane from stupid controls. <3 Will
			SDL_WarpMouse( static_cast<unsigned short>(GRAPHICS.getDimensions().getX()*.5f)
				, static_cast<unsigned short>(GRAPHICS.getDimensions().getY()*.5f) );
			SDL_PollEvent(input);
		}
	}

	handler.update(cam);  //preform all actions based on the falgs set
}

const Math_::Vector3 &InputHandler::getJohnnyMovement() const
{
	return johnnyMove;
}

}
//END OF Input_
